@@ -128,7 +128,7 @@ module Test
       #   @return [void]
       def assert(object=NOT_SPECIFIED, message=nil, &block)
         _wrap_assertion do
-          have_object = (object != NOT_SPECIFIED)
+          have_object = !NOT_SPECIFIED.equal?(object)
           if block
             message = object if have_object
             if defined?(PowerAssert)
@@ -178,7 +178,7 @@ module Test
       #   refute(false)    # => pass
       #   refute(nil)      # => pass
       #
-      # @example Fialure patterns
+      # @example Failure patterns
       #   refute(true)     # => failure
       #   refute("string") # => failure
       #
@@ -525,7 +525,7 @@ EOT
       alias_method :refute_respond_to, :assert_not_respond_to
 
       ##
-      # Passes if +string+ =~ +pattern+.
+      # Passes if +pattern+ =~ +string+.
       #
       # @example
       #   assert_match(/\d+/, 'five, 6, seven')
@@ -537,8 +537,9 @@ EOT
             else
               pattern
           end
-          full_message = build_message(message, "<?> expected to be =~\n<?>.", string, pattern)
-          assert_block(full_message) { string =~ pattern }
+          full_message = build_message(message, "<?> expected to be =~\n<?>.",
+                                       pattern, string)
+          assert_block(full_message) { pattern =~ string }
         end
       end
 
@@ -738,11 +739,47 @@ EOT
       end
 
       # @private
-      UncaughtThrow = {
-        NameError => /^uncaught throw `(.+)'$/,
-        ArgumentError => /^uncaught throw (`.+'|.+)$/,
-        ThreadError => /^uncaught throw `(.+)' in thread /,
-      }
+      class ThrowTagExtractor
+        @@have_uncaught_throw_error = const_defined?(:UncaughtThrowError)
+
+        UncaughtThrowPatterns = {
+          NameError => /^uncaught throw `(.+)'$/,
+          ArgumentError => /^uncaught throw (`.+'|.+)$/,
+          ThreadError => /^uncaught throw `(.+)' in thread /,
+        }
+
+        def initialize(error)
+          @error = error
+        end
+
+        def extract_tag
+          tag = nil
+          if @@have_uncaught_throw_error
+            return nil unless @error.is_a?(UncaughtThrowError)
+            tag = @error.tag
+          else
+            pattern = UncaughtThrowPatterns[@error.class]
+            return nil if pattern.nil?
+            return nil unless pattern =~ @error.message
+            tag = $1
+          end
+          normalize_tag(tag)
+        end
+
+        private
+        def normalize_tag(tag)
+          case tag
+          when /\A:/
+            tag[1..-1].intern
+          when /\A`(.+)'\z/
+            $1.intern
+          when String
+            tag.intern
+          else
+            tag
+          end
+        end
+      end
 
       ##
       # Passes if the block throws +expected_object+
@@ -772,9 +809,10 @@ EOT
                                          "<?> should have been thrown.",
                                          expected_object)
             assert_block(full_message) {caught}
-          rescue NameError, ArgumentError, ThreadError => error
-            raise unless UncaughtThrow[error.class] =~ error.message
-            tag = AssertionMessage.normalize_tag($1)
+          rescue => error
+            extractor = ThrowTagExtractor.new(error)
+            tag = extractor.extract_tag
+            raise if tag.nil?
             full_message = build_message(message,
                                          "<?> expected to be thrown but\n" +
                                          "<?> was thrown.",
@@ -801,9 +839,10 @@ EOT
           assert(block_given?, "Should have passed a block to assert_nothing_thrown")
           begin
             proc.call
-          rescue NameError, ArgumentError, ThreadError => error
-            raise unless UncaughtThrow[error.class] =~ error.message
-            tag = AssertionMessage.normalize_tag($1)
+          rescue => error
+            extractor = ThrowTagExtractor.new(error)
+            tag = extractor.extract_tag
+            raise if tag.nil?
             full_message = build_message(message,
                                          "<?> was thrown when nothing was expected",
                                          tag)
@@ -1670,19 +1709,6 @@ EOT
             MaybeContainer.new(value, &formatter)
           end
 
-          def normalize_tag(tag)
-            case tag
-            when /\A:/
-              tag[1..-1].intern
-            when /\A`(.+)'\z/
-              $1.intern
-            when String
-              tag.intern
-            else
-              tag
-            end
-          end
-
           MAX_DIFF_TARGET_STRING_SIZE = 1000
           def max_diff_target_string_size
             return @@max_diff_target_string_size if @@max_diff_target_string_size
@@ -2062,7 +2088,9 @@ EOT
         def to_s
           message_parts = []
           if (@head)
-            head = @head.to_s
+            head = @head
+            head = head.call if head.respond_to?(:call)
+            head = head.to_s
             unless(head.empty?)
               message_parts << add_period(head)
             end
