@@ -1,7 +1,10 @@
-require 'test/unit/color-scheme'
-require 'test/unit/priority'
-require 'test/unit/attribute-matcher'
-require 'optparse'
+require "English"
+
+require "test/unit/color-scheme"
+require "test/unit/priority"
+require "test/unit/attribute-matcher"
+require "test/unit/testcase"
+require "optparse"
 
 module Test
   module Unit
@@ -78,25 +81,30 @@ module Test
       end
 
       register_collector(:descendant) do |auto_runner|
-        require 'test/unit/collector/descendant'
+        require "test/unit/collector/descendant"
         collector = Collector::Descendant.new
         collector.filter = auto_runner.filters
-        collector.collect($0.sub(/\.rb\Z/, ''))
+        collector.collect($0.sub(/\.rb\Z/, ""))
       end
 
       register_collector(:load) do |auto_runner|
-        require 'test/unit/collector/load'
+        require "test/unit/collector/load"
         collector = Collector::Load.new
-        collector.patterns.concat(auto_runner.pattern) if auto_runner.pattern
-        collector.excludes.concat(auto_runner.exclude) if auto_runner.exclude
+        unless auto_runner.pattern.empty?
+          collector.patterns.replace(auto_runner.pattern)
+        end
+        unless auto_runner.exclude.empty?
+          collector.excludes.replace(auto_runner.exclude)
+        end
         collector.base = auto_runner.base
+        collector.default_test_paths = auto_runner.default_test_paths
         collector.filter = auto_runner.filters
         collector.collect(*auto_runner.to_run)
       end
 
       # JUST TEST!
       # register_collector(:xml) do |auto_runner|
-      #   require 'test/unit/collector/xml'
+      #   require "test/unit/collector/xml"
       #   collector = Collector::XML.new
       #   collector.filter = auto_runner.filters
       #   collector.collect(auto_runner.to_run[0])
@@ -104,26 +112,32 @@ module Test
 
       # deprecated
       register_collector(:object_space) do |auto_runner|
-        require 'test/unit/collector/objectspace'
+        require "test/unit/collector/objectspace"
         c = Collector::ObjectSpace.new
         c.filter = auto_runner.filters
-        c.collect($0.sub(/\.rb\Z/, ''))
+        c.collect($0.sub(/\.rb\Z/, ""))
       end
 
       # deprecated
       register_collector(:dir) do |auto_runner|
-        require 'test/unit/collector/dir'
+        require "test/unit/collector/dir"
         c = Collector::Dir.new
         c.filter = auto_runner.filters
-        c.pattern.concat(auto_runner.pattern) if auto_runner.pattern
-        c.exclude.concat(auto_runner.exclude) if auto_runner.exclude
+        unless auto_runner.pattern.empty?
+          c.pattern.replace(auto_runner.pattern)
+        end
+        unless auto_runner.exclude.empty?
+          c.exclude.replace(auto_runner.exclude)
+        end
         c.base = auto_runner.base
         $:.push(auto_runner.base) if auto_runner.base
-        c.collect(*(auto_runner.to_run.empty? ? ['.'] : auto_runner.to_run))
+        c.collect(*(auto_runner.to_run.empty? ? ["."] : auto_runner.to_run))
       end
 
       attr_reader :suite, :runner_options
-      attr_accessor :filters, :to_run, :pattern, :exclude, :base, :workdir
+      attr_accessor :filters, :to_run
+      attr_accessor :default_test_paths
+      attr_accessor :pattern, :exclude, :base, :workdir
       attr_accessor :color_scheme, :listeners
       attr_writer :runner, :collector
 
@@ -133,6 +147,7 @@ module Test
         @collector = default_collector
         @filters = []
         @to_run = []
+        @default_test_paths = []
         @color_scheme = ColorScheme.default
         @runner_options = {}
         @default_arguments = []
@@ -156,7 +171,7 @@ module Test
       def process_args(args=ARGV)
         begin
           args.unshift(*@default_arguments)
-          options.order!(args) {|arg| @to_run << arg}
+          options.order!(args) {|arg| add_test_path(arg)}
         rescue OptionParser::ParseError => e
           puts e
           puts options
@@ -170,49 +185,58 @@ module Test
           o.banner = "Test::Unit automatic runner."
           o.banner << "\nUsage: #{$0} [options] [-- untouched arguments]"
 
-          o.on('-r', '--runner=RUNNER', RUNNERS,
+          o.on("-r", "--runner=RUNNER", RUNNERS,
                "Use the given RUNNER.",
                "(" + keyword_display(RUNNERS) + ")") do |r|
             @runner = r
           end
 
-          o.on('--collector=COLLECTOR', COLLECTORS,
+          o.on("--collector=COLLECTOR", COLLECTORS,
                "Use the given COLLECTOR.",
                "(" + keyword_display(COLLECTORS) + ")") do |collector|
             @collector = collector
           end
 
           if (@standalone)
-            o.on('-b', '--basedir=DIR', "Base directory of test suites.") do |b|
+            o.on("-b", "--basedir=DIR", "Base directory of test suites.") do |b|
               @base = b
             end
 
-            o.on('-w', '--workdir=DIR', "Working directory to run tests.") do |w|
+            o.on("-w", "--workdir=DIR", "Working directory to run tests.") do |w|
               @workdir = w
             end
 
-            o.on('-a', '--add=TORUN', Array,
+            o.on("--default-test-path=PATH",
+                 "Add PATH to the default test paths.",
+                 "The PATH is used when user doesn't specify any test path.",
+                 "You can specify this option multiple times.") do |path|
+              @default_test_paths << path
+            end
+
+            o.on("-a", "--add=TORUN", Array,
                  "Add TORUN to the list of things to run;",
-                 "can be a file or a directory.") do |a|
-              @to_run.concat(a)
+                 "can be a file or a directory.") do |paths|
+              paths.each do |path|
+                add_test_path(path)
+              end
             end
 
             @pattern = []
-            o.on('-p', '--pattern=PATTERN', Regexp,
+            o.on("-p", "--pattern=PATTERN", Regexp,
                  "Match files to collect against PATTERN.") do |e|
               @pattern << e
             end
 
             @exclude = []
-            o.on('-x', '--exclude=PATTERN', Regexp,
+            o.on("-x", "--exclude=PATTERN", Regexp,
                  "Ignore files to collect against PATTERN.") do |e|
               @exclude << e
             end
           end
 
-          o.on('-n', '--name=NAME', String,
+          o.on("-n", "--name=NAME", String,
                "Runs tests matching NAME.",
-               "(patterns may be used).") do |name|
+               "Use '/PATTERN/' for NAME to use regular expression.") do |name|
             name = (%r{\A/(.*)/\Z} =~ name ? Regexp.new($1) : name)
             @filters << lambda do |test|
               return true if name === test.method_name
@@ -224,9 +248,9 @@ module Test
             end
           end
 
-          o.on('--ignore-name=NAME', String,
+          o.on("--ignore-name=NAME", String,
                "Ignores tests matching NAME.",
-               "(patterns may be used).") do |n|
+               "Use '/PATTERN/' for NAME to use regular expression.") do |n|
             n = (%r{\A/(.*)/\Z} =~ n ? Regexp.new($1) : n)
             case n
             when Regexp
@@ -236,25 +260,25 @@ module Test
             end
           end
 
-          o.on('-t', '--testcase=TESTCASE', String,
+          o.on("-t", "--testcase=TESTCASE", String,
                "Runs tests in TestCases matching TESTCASE.",
-               "(patterns may be used).") do |n|
+               "Use '/PATTERN/' for TESTCASE to use regular expression.") do |n|
             n = (%r{\A/(.*)/\Z} =~ n ? Regexp.new($1) : n)
             @filters << lambda do |test|
               match_test_case_name(test, n)
             end
           end
 
-          o.on('--ignore-testcase=TESTCASE', String,
+          o.on("--ignore-testcase=TESTCASE", String,
                "Ignores tests in TestCases matching TESTCASE.",
-               "(patterns may be used).") do |n|
+               "Use '/PATTERN/' for TESTCASE to use regular expression.") do |n|
             n = (%r{\A/(.*)/\Z} =~ n ? Regexp.new($1) : n)
             @filters << lambda do |test|
               not match_test_case_name(test, n)
             end
           end
 
-          o.on('--location=LOCATION', String,
+          o.on("--location=LOCATION", String,
                "Runs tests that defined in LOCATION.",
                "LOCATION is one of PATH:LINE, PATH or LINE") do |location|
             if /\A\d+\z/ =~ location
@@ -264,14 +288,10 @@ module Test
               path, line, = location.split(/:(\d+)/, 2)
               line = line.to_i unless line.nil?
             end
-            @filters << lambda do |test|
-              test.class.test_defined?(:path => path,
-                                       :line => line,
-                                       :method_name => test.method_name)
-            end
+            add_location_filter(path, line)
           end
 
-          o.on('--attribute=EXPRESSION', String,
+          o.on("--attribute=EXPRESSION", String,
                "Runs tests that matches EXPRESSION.",
                "EXPRESSION is evaluated as Ruby's expression.",
                "Test attribute name can be used with no receiver in EXPRESSION.",
@@ -309,7 +329,7 @@ module Test
             Priority.default = priority
           end
 
-          o.on('-I', "--load-path=DIR[#{File::PATH_SEPARATOR}DIR...]",
+          o.on("-I", "--load-path=DIR[#{File::PATH_SEPARATOR}DIR...]",
                "Appends directory list to $LOAD_PATH.") do |dirs|
             $LOAD_PATH.concat(dirs.split(File::PATH_SEPARATOR))
           end
@@ -345,23 +365,23 @@ module Test
             option_builder.call(self, o)
           end
 
-          o.on('--',
+          o.on("--",
                "Stop processing options so that the",
                "remaining options will be passed to the",
                "test."){o.terminate}
 
-          o.on('-h', '--help', 'Display this help.'){puts o; exit}
+          o.on("-h", "--help", "Display this help."){puts o; exit}
 
           o.on_tail
-          o.on_tail('Deprecated options:')
+          o.on_tail("Deprecated options:")
 
-          o.on_tail('--console', 'Console runner (use --runner).') do
+          o.on_tail("--console", "Console runner (use --runner).") do
             warn("Deprecated option (--console).")
             @runner = self.class.runner(:console)
           end
 
           if RUNNERS[:fox]
-            o.on_tail('--fox', 'Fox runner (use --runner).') do
+            o.on_tail("--fox", "Fox runner (use --runner).") do
               warn("Deprecated option (--fox).")
               @runner = self.class.runner(:fox)
             end
@@ -405,7 +425,7 @@ module Test
       end
 
       def load_config(file)
-        require 'yaml'
+        require "yaml"
         config = YAML.load(File.read(file))
         runner_name = config["runner"]
         @runner = self.class.runner(runner_name) || @runner
@@ -467,10 +487,27 @@ module Test
         end
         false
       end
+
+      def add_test_path(path)
+        if /:(\d+)\z/ =~ path
+          line = $1.to_i
+          path = $PREMATCH
+          add_location_filter(path, line)
+        end
+        @to_run << path
+      end
+
+      def add_location_filter(path, line)
+        @filters << lambda do |test|
+          test.class.test_defined?(:path => path,
+                                   :line => line,
+                                   :method_name => test.method_name)
+        end
+      end
     end
   end
 end
 
-require 'test/unit/runner/console'
-require 'test/unit/runner/emacs'
-require 'test/unit/runner/xml'
+require "test/unit/runner/console"
+require "test/unit/runner/emacs"
+require "test/unit/runner/xml"
