@@ -20,6 +20,7 @@ require 'test/unit/data'
 require 'test/unit/testsuite'
 require 'test/unit/test-suite-creator'
 require 'test/unit/assertion-failed-error'
+require 'test/unit/auto-runner-loader'
 require 'test/unit/util/backtracefilter'
 require 'test/unit/util/output'
 require 'test/unit/util/method-owner-finder'
@@ -104,7 +105,6 @@ module Test
 
       class << self
         def inherited(sub_class) # :nodoc:
-          require "test/unit"
           DESCENDANTS << sub_class
           super
         end
@@ -120,16 +120,22 @@ module Test
           source_location = find_attribute(stringified_name, :source_location)
           if source_location
             path, line = source_location
+          elsif respond_to?(:caller_locations, true)
+            location = caller_locations(1, 1)[0]
+            path = location.absolute_path || location.path
+            line = location.lineno
           else
-            path, line, = caller[0].split(/:(\d+)/,2)
+            # TODO: Remove me when Ruby 1.9 support is dropped
+            path, line, = caller[0].split(/:(\d+)/, 2)
             line = line.to_i if line
           end
           method_locations << {
             :method_name => stringified_name,
-            :path => path,
+            :path => File.expand_path(path),
             :line => line,
           }
           added_method_names[stringified_name] = true
+          AutoRunnerLoader.check(self, stringified_name)
         end
 
         def added_method_names # :nodoc:
@@ -281,6 +287,9 @@ module Test
           else
             targets = test_description_or_targets
             attribute(:test, true, {}, *targets)
+            targets.each do |target|
+              AutoRunnerLoader.check(self, target)
+            end
           end
         end
 
@@ -453,14 +462,31 @@ module Test
           @internal_data.test_started
           yield(STARTED, name)
           yield(STARTED_OBJECT, self)
+          processed_exception_in_setup = false
           begin
-            run_setup
-            run_test
-            run_cleanup
-            add_pass
+            catch do |tag|
+              run_setup do
+                begin
+                  run_test
+                  run_cleanup
+                  add_pass
+                rescue Exception
+                  @internal_data.interrupted
+                  unless handle_exception($!)
+                    processed_exception_in_setup = true
+                    raise
+                  end
+                  throw(tag)
+                end
+              end
+            end
           rescue Exception
-            @internal_data.interrupted
-            raise unless handle_exception($!)
+            if processed_exception_in_setup
+              raise
+            else
+              @internal_data.interrupted
+              raise unless handle_exception($!)
+            end
           ensure
             begin
               run_teardown
