@@ -1,5 +1,5 @@
 # Author:: Nathaniel Talbott.
-# Copyright:: Copyright (c) 2008-2013  Kouhei Sutou <kou@clear-code.com>
+# Copyright:: Copyright (c) 2008-2014  Kouhei Sutou <kou@clear-code.com>
 # Copyright:: Copyright (c) 2011  Haruka Yoshihara <yoshihara@clear-code.com>
 # Copyright:: Copyright (c) 2000-2002  Nathaniel Talbott
 # License:: Ruby license.
@@ -28,7 +28,9 @@ module Test
       def setup
         @tc_failure_error = Class.new(TestCase) do
           def test_failure
-            assert_block("failure") { false }
+            assert_block("failure") do
+              false
+            end
           end
           def test_error
             1 / 0
@@ -37,7 +39,9 @@ module Test
             nested
           end
           def nested
-            assert_block("nested"){false}
+            assert_block("nested") do
+              false
+            end
           end
           def return_passed?
             return passed?
@@ -49,82 +53,170 @@ module Test
         end
       end
 
+      def jruby_backtrace_entry?(entry)
+        entry.start_with?("org/jruby/")
+      end
+
+      def rubinius_backtrace_entry?(entry)
+        entry.start_with?("kernel/")
+      end
+
+      def normalize_location(location)
+        filtered_location = location.reject do |entry|
+          jruby_backtrace_entry?(entry) or
+            rubinius_backtrace_entry?(entry)
+        end
+        filtered_location.collect do |entry|
+          entry.sub(/:\d+:/, ":0:")
+        end
+      end
+
       def test_add_failed_assertion
         test_case = @tc_failure_error.new(:test_failure)
-        check("passed? should start out true", test_case.return_passed?)
-        result = TestResult.new
-        called = false
-        result.add_listener(TestResult::FAULT) {
-          | fault |
-          check("Should have a Failure", fault.instance_of?(Failure))
-          check("The Failure should have the correct message", "failure" == fault.message)
-          check("The Failure should have the correct test_name (was <#{fault.test_name}>)", fault.test_name == "test_failure(TC_FailureError)")
-          r = /\A.*#{Regexp.escape(File.basename(__FILE__))}:\d+:in `test_failure'\Z/
+        assert do
+          test_case.passed?
+        end
 
-          location = fault.location
-          check("The location should be an array", location.kind_of?(Array))
-          check("The location should have two lines (was: <#{location.inspect}>)", location.size == 2)
-          check("The Failure should have the correct location (was <#{location[0].inspect}>, expected <#{r.inspect}>)", r =~ location[0])
-          called = true
-        }
+        result = TestResult.new
+        faults = []
+        result.add_listener(TestResult::FAULT) do |fault|
+          faults << fault
+        end
         progress = []
-        test_case.run(result) { |*arguments| progress << arguments }
-        check("The failure should have triggered the listener", called)
-        check("The failure should have set passed?", !test_case.return_passed?)
-        check("The progress block should have been updated correctly",
-              [[TestCase::STARTED, test_case.name],
-               [TestCase::STARTED_OBJECT, test_case],
-               [TestCase::FINISHED, test_case.name],
-               [TestCase::FINISHED_OBJECT, test_case]] == progress)
+        test_case.run(result) do |*arguments|
+          progress << arguments
+        end
+        fault_details = faults.collect do |fault|
+          {
+            :class     => fault.class,
+            :message   => fault.message,
+            :test_name => fault.test_name,
+            :location  => normalize_location(fault.location),
+          }
+        end
+        assert_equal([
+                       {
+                         :class     => Failure,
+                         :message   => "failure",
+                         :test_name => "test_failure(TC_FailureError)",
+                         :location  => [
+                           "#{__FILE__}:0:in `test_failure'",
+                           "#{__FILE__}:0:in `#{__method__}'",
+                         ],
+                       },
+                     ],
+                     fault_details)
+
+        assert do
+          not test_case.passed?
+        end
+        assert_equal([
+                       [TestCase::STARTED,         test_case.name],
+                       [TestCase::STARTED_OBJECT,  test_case],
+                       [TestCase::FINISHED,        test_case.name],
+                       [TestCase::FINISHED_OBJECT, test_case],
+                     ],
+                     progress)
       end
 
       def test_add_failure_nested
         test_case = @tc_failure_error.new(:test_nested_failure)
-        check("passed? should start out true", test_case.return_passed?)
+        assert do
+          test_case.passed?
+        end
 
         result = TestResult.new
-        called = false
-        result.add_listener(TestResult::FAULT) {
-          | fault |
-          check("Should have a Failure", fault.instance_of?(Failure))
-          check("The Failure should have the correct message", "nested" == fault.message)
-          check("The Failure should have the correct test_name (was <#{fault.test_name}>)", fault.test_name == "test_nested_failure(TC_FailureError)")
+        faults = []
+        result.add_listener(TestResult::FAULT) do |fault|
+          faults << fault
+        end
+        test_case.run(result) do
+        end
+        fault_details = faults.collect do |fault|
+          {
+            :class     => fault.class,
+            :message   => fault.message,
+            :test_name => fault.test_name,
+            :location  => normalize_location(fault.location),
+          }
+        end
+        assert_equal([
+                       {
+                         :class     => Failure,
+                         :message   => "nested",
+                         :test_name => "test_nested_failure(TC_FailureError)",
+                         :location  => [
+                           "#{__FILE__}:0:in `nested'",
+                           "#{__FILE__}:0:in `test_nested_failure'",
+                           "#{__FILE__}:0:in `#{__method__}'",
+                         ],
+                       },
+                     ],
+                     fault_details)
 
-          location = fault.location
-          check("The location should be an array", location.kind_of?(Array))
-          check("The location should have the correct number of lines (was: <#{location.inspect}>)", location.size == 3)
-          check("The Failure should have the correct location (was <#{location[0].inspect}>)", /\A.*#{Regexp.escape(File.basename(__FILE__))}:\d+:in `nested'\Z/ =~ location[0])
-          check("The Failure should have the correct location (was <#{location[1].inspect}>)", /\A.*#{Regexp.escape(File.basename(__FILE__))}:\d+:in `test_nested_failure'\Z/ =~ location[1])
-          called = true
-        }
-        test_case.run(result){}
-        check("The failure should have triggered the listener", called)
+        assert do
+          not test_case.passed?
+        end
       end
-      
+
+      def jruby?
+        defined?(JRUBY_VERSION)
+      end
+
+      def rubinius?
+        false # TODO
+      end
+
+      def cruby?
+        (not jruby?) and (not rubinius?)
+      end
+
       def test_add_error
         test_case = @tc_failure_error.new(:test_error)
-        check("passed? should start out true", test_case.return_passed?)
+        assert do
+          test_case.passed?
+        end
+
         result = TestResult.new
-        called = false
-        result.add_listener(TestResult::FAULT) {
-          | fault |
-          check("Should have a TestError", fault.instance_of?(Error))
-          check("The Error should have the correct message", "ZeroDivisionError: divided by 0" == fault.message)
-          check("The Error should have the correct test_name", "test_error(TC_FailureError)" == fault.test_name)
-          check("The Error should have the correct exception", fault.exception.instance_of?(ZeroDivisionError))
-          called = true
-        }
-        test_case.run(result) {}
-        check("The error should have triggered the listener", called)
-        check("The error should have set passed?", !test_case.return_passed?)
+        faults = []
+        result.add_listener(TestResult::FAULT) do |fault|
+          faults << fault
+        end
+        test_case.run(result) do
+        end
+        fault_details = faults.collect do |fault|
+          {
+            :class     => fault.class,
+            :message   => fault.message,
+            :test_name => fault.test_name,
+            :location  => normalize_location(fault.location),
+          }
+        end
+        location = []
+        location << "#{__FILE__}:0:in `/'" if cruby?
+        location << "#{__FILE__}:0:in `test_error'"
+        location << "#{__FILE__}:0:in `#{__method__}'"
+        assert_equal([
+                       {
+                         :class     => Error,
+                         :message   => "ZeroDivisionError: divided by 0",
+                         :test_name => "test_error(TC_FailureError)",
+                         :location  => location,
+                       },
+                     ],
+                     fault_details)
+
+        assert do
+          not test_case.passed?
+        end
       end
 
-      def test_no_tests      
+      def test_no_tests
         suite = TestCase.suite
         check("Should have a test suite", suite.instance_of?(TestSuite))
         check("Should have one test", suite.size == 1)
         check("Should have the default test", suite.tests.first.name == "default_test(Test::Unit::TestCase)")
-        
+
         result = TestResult.new
         suite.run(result) {}
         check("Should have had one test run", result.run_count == 1)
@@ -153,19 +245,18 @@ module Test
             assert_block {true}
           end
         end
-      
+
         suite = tc.suite
         check("Should have a test suite", suite.instance_of?(TestSuite))
         check("Should have three tests", suite.size == 3)
-  
+
         result = TestResult.new
         suite.run(result) {}
         check("Should have had three test runs", result.run_count == 3)
         check("Should have had one test failure", result.failure_count == 1)
         check("Should have had one test error", result.error_count == 1)
       end
-      
-     
+
       def test_setup_teardown
         tc = Class.new(TestCase) do
           attr_reader(:setup_called, :teardown_called)
@@ -191,34 +282,34 @@ module Test
           end
         end
         result = TestResult.new
-  
+
         test = tc.new(:test_succeed)
         test.run(result) {}
         check("Should have called setup the correct number of times", test.setup_called)
         check("Should have called teardown the correct number of times", test.teardown_called)
-  
+
         test = tc.new(:test_fail)
         test.run(result) {}
         check("Should have called setup the correct number of times", test.setup_called)
         check("Should have called teardown the correct number of times", test.teardown_called)
-  
+
         test = tc.new(:test_error)
         test.run(result) {}
         check("Should have called setup the correct number of times", test.setup_called)
         check("Should have called teardown the correct number of times", test.teardown_called)
-  
+
         check("Should have had two test runs", result.run_count == 3)
         check("Should have had a test failure", result.failure_count == 1)
         check("Should have had a test error", result.error_count == 1)
       end
-      
+
       def test_assertion_failed_not_called
         tc = Class.new(TestCase) do
           def test_thing
             raise AssertionFailedError.new
           end
         end
-        
+
         suite = tc.suite
         check("Should have one test", suite.size == 1)
         result = TestResult.new
@@ -227,7 +318,7 @@ module Test
         check("Should have had one assertion failure", result.failure_count == 1)
         check("Should not have any assertion errors but had #{result.error_count}", result.error_count == 0)
       end
-      
+
       def test_equality
         tc1 = Class.new(TestCase) do
           def test_1
@@ -235,31 +326,30 @@ module Test
           def test_2
           end
         end
-        
+
         tc2 = Class.new(TestCase) do
           def test_1
           end
         end
-      
+
         test1 = tc1.new('test_1')
         test2 = tc1.new('test_1')
         check("Should be equal", test1 == test2)
         check("Should be equal", test2 == test1)
-        
+
         test1 = tc1.new('test_2')
         check("Should not be equal", test1 != test2)
         check("Should not be equal", test2 != test1)
-        
+
         test2 = tc1.new('test_2')
         check("Should be equal", test1 == test2)
         check("Should be equal", test2 == test1)
-        
+
         test1 = tc1.new('test_1')
         test2 = tc2.new('test_1')
         check("Should not be equal", test1 != test2)
         check("Should not be equal", test2 != test1)
 
-        
         check("Should not be equal", test1 != Object.new)
         check("Should not be equal", Object.new != test1)
       end
@@ -570,6 +660,28 @@ module Test
               end
               assert_true(test_case.test_defined?(:line => line_after))
             end
+
+            def test_child
+              child_test_case = nil
+              line_child = nil
+              parent_test_case = Class.new(TestCase) do
+                test "parent" do
+                end
+
+                child_test_case = Class.new(self) do
+                  line_child = __LINE__; test "child" do
+                  end
+                end
+              end
+              assert_equal([
+                             false,
+                             true,
+                           ],
+                           [
+                             parent_test_case.test_defined?(:line => line_child),
+                             child_test_case.test_defined?(:line => line_child),
+                           ])
+            end
           end
 
           class TestMethodStyle < self
@@ -601,6 +713,28 @@ module Test
               end
               assert_true(test_case.test_defined?(:line => line_after))
             end
+
+            def test_child
+              child_test_case = nil
+              line_child = nil
+              parent_test_case = Class.new(TestCase) do
+                test "parent" do
+                end
+
+                child_test_case = Class.new(self) do
+                  line_child = __LINE__; test "child" do
+                  end
+                end
+              end
+              assert_equal([
+                             false,
+                             true,
+                           ],
+                           [
+                             parent_test_case.test_defined?(:line => line_child),
+                             child_test_case.test_defined?(:line => line_child),
+                           ])
+            end
           end
         end
 
@@ -752,6 +886,33 @@ module Test
           end
           assert_equal(["test_nothing"], test_method_names)
         end
+
+        def test_duplicated_name
+          test_case = Class.new(TestCase) do
+            def test_nothing
+            end
+          end
+          sub_test_case = test_case.sub_test_case("sub test case") do
+            def test_nothing
+            end
+          end
+
+          test_method_names = test_case.suite.tests.collect do |test|
+            test.method_name
+          end
+          sub_test_method_names = sub_test_case.suite.tests.collect do |test|
+            test.method_name
+          end
+
+          assert_equal([
+                         ["test_nothing"],
+                         ["test_nothing"],
+                       ],
+                       [
+                         test_method_names,
+                         sub_test_method_names,
+                       ])
+        end
       end
 
       class TestStartupShutdown < self
@@ -974,6 +1135,35 @@ module Test
               run_test_case(test_case)
             end
           end
+
+          def test_pass_through_in_test
+            test_case = Class.new(TestCase) do
+              @called = []
+              class << self
+                def called
+                  @called
+                end
+
+                def startup
+                  @called << :startup
+                end
+
+                def shutdown
+                  @called << :shutdown
+                end
+              end
+
+              def test_error
+                raise Interrupt, "from test"
+              end
+            end
+
+            assert_raise(Interrupt) do
+              run_test_case(test_case)
+            end
+            assert_equal([:startup, :shutdown],
+                         test_case.called)
+          end
         end
       end
     end
